==> videosdk-0.0.7-py3-none-any.whl: videosdk/participant.py <==
import asyncio
import base64
from io import BytesIO
import logging
from typing import Any, Callable, Dict, Optional, TypedDict
from PIL import Image
from ._events import Events, VideoSDKEvents
from .event_handler import BaseEvent
from .lib.pymediasoup.consumer import Consumer
from .lib.pymediasoup.producer import Producer
from .room_client import RoomClient
from .stream import Stream

class ParticipantConfig(TypedDict, total=False):
    id: str
    display_name: str
    local: bool
    mode: str
    meta_data: Optional[Dict[str, Any]]
    room_client: RoomClient


class Participant:
    """
    Represents a participant in a video conference.

    Attributes:
        id (str): The participant's unique identifier.
        display_name (str): The participant's display name.
        local (bool): Flag indicating if the participant is local.
        mode (str): The mode of the participant.
        meta_data (dict, optional): Additional metadata about the participant.
        room_client (RoomClient): The client used to interact with the room.
        streams (dict[str, Stream]): Dictionary of streams associated with the participant.
        mic_on (bool): Flag indicating if the participant's microphone is on.
        webcam_on (bool): Flag indicating if the participant's webcam is on.
        listeners (list[BaseEvent]): List of event listeners registered for the participant.
    """

    def __init__(
        self,
        id: str,
        display_name: str = "",
        local: bool = False,
        mode: str = "",
        meta_data: Any = None,
        room_client: RoomClient = None,
    ):
        """
        Initialize a Participant instance.

        Args:
            id (str): The participant's unique identifier.
            display_name (str, optional): The participant's display name. Default is "".
            local (bool, optional): Flag indicating if the participant is local. Default is False.
            mode (str, optional): The mode of the participant. Default is "".
            meta_data (Any, optional): Additional metadata about the participant. Default is None.
            room_client (RoomClient, optional): The client used to interact with the room. Default is None.
        """
        self.__id = id
        self.__display_name = display_name
        self.__local = local
        self.__mode = mode
        self.__meta_data = meta_data
        self.__room_client = room_client
        self.__streams: dict[str, Stream] = {}
        self.__mic_on: bool = False
        self.__webcam_on: bool = False
        self.__listeners: list[BaseEvent] = []
        self.__internal_event_listeners: dict[str, Callable] = {}
        self.__tasks: dict[str, asyncio.Task] = {}
        self.__listen()
        logging.debug(f"Meeting Participant initialized with ID: {id}")

    def __del__(self):
        logging.debug("Meeting :: participant deleted %s", self.__id)

    def __emit(self, event_name, data=None):
        """
        Emit an event to all registered listeners.

        Args:
            event_name (str): The name of the event to emit.
            data (Any, optional): Optional data associated with the event. Default is None.
        """
        for _listener in self.__listeners:
            logging.debug(f"{self.id} :: Emitting Event: {event_name}")
            _listener.handle_event(event_name, data)

    def add_event_listener(self, event_handler: BaseEvent):
        """
        Add an event listener for this participant.

        Args:
            event_handler (BaseEvent): The event handler object to add.
        """
        self.__listeners.append(event_handler)

    def remove_event_listener(self, event_handler: BaseEvent):
        """
        Remove an event listener from this participant.

        Args:
            event_handler (BaseEvent): The event handler object to remove.
        """
        self.__listeners.remove(event_handler)

    def remove_all_listeners(self):
        """
        Remove all event listeners registered for this participant.
        """
        logging.debug("Removing all event listeners for %s", self.__id)
        self.__listeners.clear()

    def _cleanup(self):
        """
        Clean up internal resources and event listeners.
        """
        for k, v in self.__internal_event_listeners.items():
            self.__room_client.remove_listener(k, v)

        for task in self.__tasks.values():
            task.cancel()

    def __listen(self):
        """
        Register internal event listeners for room events.
        """
        @self.__room_client.on(Events.ADD_CONSUMER)
        def on_stream_enabled(consumer: Consumer):
            """
            Handle event when a stream is enabled for this participant.

            Args:
                consumer (Consumer): The consumer object representing the stream.
            """
            if consumer.appData["peerId"] == self.__id:
                stream = Stream(track=consumer.track)
                self.__streams[consumer.id] = stream
                logging.debug(
                    f"Participant :: {self.id} :: stream-enabled {stream.kind}")
                self.__emit(VideoSDKEvents.EV_STREAM_ENABLED.value, stream)

        self.__internal_event_listeners[Events.ADD_CONSUMER] = on_stream_enabled

        @self.__room_client.on(Events.REMOVE_CONSUMER)
        def on_stream_disabled(consumer: Consumer):
            """
            Handle event when a stream is disabled for this participant.

            Args:
                consumer (Consumer): The consumer object representing the stream.
            """
            if consumer.appData["peerId"] == self.__id:
                if consumer.id in self.__streams:
                    stream = self.__streams.pop(consumer.id)
                    stream.track.stop()
                    logging.debug(
                        f"Participant :: {self.id} :: stream-disabled {stream.kind}")
                    self.__emit(
                        VideoSDKEvents.EV_STREAM_DISABLED.value, stream)

        self.__internal_event_listeners[Events.REMOVE_CONSUMER] = on_stream_disabled

        @self.__room_client.on(Events.PARTICIPANT_MEDIA_STATE_CHANGED)
        def on_media_status_changed(data):
            """
            Handle event when the media status (audio/video) changes for this participant.

            Args:
                data (dict): Data containing information about the media status change.
            """
            if data["peerId"] == self.__id:
                new_state = data["newState"]
                kind = data["kind"]
                if kind == "audio":
                    self.__mic_on = new_state
                elif kind == "video":
                    self.__webcam_on = new_state
                logging.debug(
                    f"Participant :: {self.id} :: media-status-changed")
                self.__emit(
                    VideoSDKEvents.EV_MEDIA_STATUS_CHANGED.value,
                    {"peerId": data["peerId"], "kind": data["kind"],
                        "newState": data["newState"]},
                )

        self.__internal_event_listeners[Events.PARTICIPANT_MEDIA_STATE_CHANGED] = on_media_status_changed

        @self.__room_client.on(Events.VIDEO_QUALITY_CHANGED)
        def on_video_quality_changed(data):
            """
            Handle event when the video quality changes for this participant.

            Args:
                data (dict): Data containing information about the video quality change.
            """
            if data["peerId"] == self.__id:
                logging.debug(
                    f"Participant :: {self.id} :: video-quality-changed")
                self.__emit(
                    VideoSDKEvents.EV_VIDEO_QUALITY_CHANGED.value,
                    {"peerId": data["peerId"], "prevQuality": data["prevQuality"],
                        "currentQuality": data["currentQuality"]},
                )

        self.__internal_event_listeners[Events.VIDEO_QUALITY_CHANGED] = on_video_quality_changed

    def remove(self):
        """
        Remove this participant from the room.
        """
        self.__room_client.remove_peer(self.__id)

    def enable_mic(self):
        """
        Enable the microphone for this participant.
        """
        self.__room_client.enable_peer_mic(self.__id)

    def disable_mic(self):
        """
        Disable the microphone for this participant.
        """
        self.__room_client.disable_peer_mic(self.__id)

    def enable_webcam(self):
        """
        Enable the webcam for this participant.
        """
        self.__room_client.enable_peer_webcam(self.__id)

    def disable_webcam(self):
        """
        Disable the webcam for this participant.
        """
        self.__room_client.disable_peer_webcam(self.__id)

    def capture_image(self, filepath="captured_image.jpg", desr_height: int = None, desr_width: int = None):
        """
        Capture an image from the participant's video stream and save it to a file.

        Args:
            filepath (str, optional): The file path where the image will be saved. Default is "captured_image.jpg".
            desr_height (int, optional): Desired height of the captured image. Default is None.
            desr_width (int, optional): Desired width of the captured image. Default is None.
        """
        self.__tasks["capture_image"] = asyncio.ensure_future(
            self.async_capture_image(filepath, desr_height, desr_width)
        )

    async def async_capture_image(self, filepath="captured_image.jpg", desr_height=None, desr_width=None) -> Image:
        """
        Asynchronously capture an image from the participant's video stream and return it as a PIL Image object.

        Args:
            filepath (str, optional): The file path where the image will be saved. Default is "captured_image.jpg".
            desr_height (int, optional): Desired height of the captured image. Default is None.
            desr_width (int, optional): Desired width of the captured image. Default is None.

        Returns:
            Image: The captured image as a PIL Image object.
        """
        try:
            streams = [stream for stream in self.__streams.values()
                       if stream.kind == "video"]

            if len(streams) == 0:
                raise Exception("Video Track Not Found")

            frame = await streams[0].track.current_frame()
            
            img = frame.to_ndarray(format="rgb24")
            
            video_height, video_width, _ = img.shape

            ratio = 16 / 9

            if desr_height is None and desr_width is None:
                desr_height = video_height
                desr_width = video_width

            if desr_height is not None:
                if desr_height <= 0 or desr_height > video_height:
                    desr_height = video_height

            if desr_width is not None:
                if desr_width <= 0 or desr_width > video_width:
                    desr_width = video_width

            if desr_height is None:
                desr_height = desr_width / ratio

            if desr_width is None:
                desr_width = desr_height * ratio

            img = Image.fromarray(img)
            resized_img = img.resize((int(desr_width), int(desr_height)))

            resized_img.save(filepath, "JPEG")
            logging.debug(f"Image captured and saved as {filepath}")

            buffered = BytesIO()

            resized_img.save(buffered, format="JPEG")

            base64_data = base64.b64encode(buffered.getvalue()).decode("utf-8")
            return base64_data

        except Exception as e:
            logging.error(f"Error in capture_image_from_stream: {e}")
            self.__room_client.emit(Events.ERROR, f"Not Able to capture image {e}")

    def _add_stream(self, producer: Producer):
        """
        Add a new stream (producer) for this participant.

        Args:
            producer (Producer): The producer object representing the stream.
        """
        stream = Stream(track=producer.track)
        self.__emit(VideoSDKEvents.EV_STREAM_ENABLED.value, stream)
        self.__streams[producer.id] = stream

    def _remove_stream(self, producer: Producer):
        """
        Remove a stream (producer) from this participant.

        Args:
            producer (Producer): The producer object representing the stream.
        """
        if producer.id in self.__streams:
            stream = self.__streams.pop(producer.id)
            self.__emit(VideoSDKEvents.EV_STREAM_DISABLED.value, stream)
            stream.track.stop()

    def _update_stream(self, producer: Producer):
        """
        Update an existing stream (producer) for this participant.

        Args:
            producer (Producer): The producer object representing the stream.
        """
        self.streams[producer.id] = producer.track

    @property
    def id(self) -> str:
        return self.__id

    @property
    def display_name(self) -> str:
        return self.__display_name

    @property
    def local(self) -> bool:
        return self.__local

    @property
    def mode(self) -> str:
        return self.__mode

    @property
    def meta_data(self) -> Any:
        return self.__meta_data

    @property
    def listeners(self) -> list[BaseEvent]:
        return self.__listeners

    @property
    def mic_on(self) -> bool:
        return self.__mic_on

    @property
    def webcam_on(self) -> bool:
        return self.__webcam_on

    @property
    def streams(self) -> dict[str, Stream]:
        return self.__streams

==> videosdk-0.0.11-py3-none-any.whl: videosdk/participant.py <==
import asyncio
import base64
from io import BytesIO
import logging
from typing import Any, Callable, Dict, Optional, TypedDict
from PIL import Image
from ._events import Events, VideoSDKEvents
from .event_handler import BaseEvent
from .lib.pymediasoup.consumer import Consumer
from .lib.pymediasoup.producer import Producer
from .room_client import RoomClient
from .stream import Stream

logger = logging.getLogger(__name__)


class ParticipantConfig(TypedDict, total=False):
    id: str
    display_name: str
    local: bool
    mode: str
    meta_data: Optional[Dict[str, Any]]
    room_client: RoomClient


class Participant:
    """
    Represents a participant in a video conference.

    Attributes:
        id (str): The participant's unique identifier.
        display_name (str): The participant's display name.
        local (bool): Flag indicating if the participant is local.
        mode (str): The mode of the participant.
        meta_data (dict, optional): Additional metadata about the participant.
        room_client (RoomClient): The client used to interact with the room.
        streams (dict[str, Stream]): Dictionary of streams associated with the participant.
        mic_on (bool): Flag indicating if the participant's microphone is on.
        webcam_on (bool): Flag indicating if the participant's webcam is on.
        listeners (list[BaseEvent]): List of event listeners registered for the participant.
    """

    def __init__(
        self,
        id: str,
        display_name: str = "",
        local: bool = False,
        mode: str = "",
        meta_data: Any = None,
        room_client: RoomClient = None,
        device: dict[str, Any] = None,
    ):
        """
        Initialize a Participant instance.

        Args:
            id (str): The participant's unique identifier.
            display_name (str, optional): The participant's display name. Default is "".
            local (bool, optional): Flag indicating if the participant is local. Default is False.
            mode (str, optional): The mode of the participant. Default is "".
            meta_data (Any, optional): Additional metadata about the participant. Default is None.
            room_client (RoomClient, optional): The client used to interact with the room. Default is None.
        """
        self.__id = id
        self.__display_name = display_name
        self.__local = local
        self.__mode = mode
        self.__meta_data = meta_data
        self.__room_client = room_client
        self.__streams: dict[str, Stream] = {}
        self.__mic_on: bool = False
        self.__webcam_on: bool = False
        self.__device: dict[str, Any] = device
        self.__listeners: list[BaseEvent] = []
        self.__internal_event_listeners: dict[str, Callable] = {}
        self.__tasks: dict[str, asyncio.Task] = {}
        self.__listen()
        logger.debug(f"Meeting Participant initialized with ID: {id}")

    def __del__(self):
        logger.debug("Meeting :: participant deleted %s", self.__id)

    def __emit(self, event_name, data=None):
        """
        Emit an event to all registered listeners.

        Args:
            event_name (str): The name of the event to emit.
            data (Any, optional): Optional data associated with the event. Default is None.
        """
        for _listener in self.__listeners:
            logger.debug(f"{self.id} :: Emitting Event: {event_name}")
            _listener.handle_event(event_name, data)

    def add_event_listener(self, event_handler: BaseEvent):
        """
        Add an event listener for this participant.

        Args:
            event_handler (BaseEvent): The event handler object to add.
        """
        self.__listeners.append(event_handler)

    def remove_event_listener(self, event_handler: BaseEvent):
        """
        Remove an event listener from this participant.

        Args:
            event_handler (BaseEvent): The event handler object to remove.
        """
        self.__listeners.remove(event_handler)

    def remove_all_listeners(self):
        """
        Remove all event listeners registered for this participant.
        """
        logger.debug("Removing all event listeners for %s", self.__id)
        self.__listeners.clear()

    def _cleanup(self):
        """
        Clean up internal resources and event listeners.
        """
        for k, v in self.__internal_event_listeners.items():
            self.__room_client.remove_listener(k, v)

        for task in self.__tasks.values():
            task.cancel()

    def __listen(self):
        """
        Register internal event listeners for room events.
        """

        @self.__room_client.on(Events.ADD_CONSUMER)
        def on_stream_enabled(consumer: Consumer):
            """
            Handle event when a stream is enabled for this participant.

            Args:
                consumer (Consumer): The consumer object representing the stream.
            """
            if consumer.appData["peerId"] == self.__id:
                stream = Stream(track=consumer.track)
                self.__streams[consumer.id] = stream
                logger.debug(
                    f"Participant :: {self.id} :: stream-enabled {stream.kind}"
                )
                self.__emit(VideoSDKEvents.EV_STREAM_ENABLED.value, stream)

        self.__internal_event_listeners[Events.ADD_CONSUMER] = on_stream_enabled

        @self.__room_client.on(Events.REMOVE_CONSUMER)
        def on_stream_disabled(consumer: Consumer):
            """
            Handle event when a stream is disabled for this participant.

            Args:
                consumer (Consumer): The consumer object representing the stream.
            """
            if consumer.appData["peerId"] == self.__id:
                if consumer.id in self.__streams:
                    stream = self.__streams.pop(consumer.id)
                    stream.track.stop()
                    logger.debug(
                        f"Participant :: {self.id} :: stream-disabled {stream.kind}"
                    )
                    self.__emit(VideoSDKEvents.EV_STREAM_DISABLED.value, stream)

        self.__internal_event_listeners[Events.REMOVE_CONSUMER] = on_stream_disabled

        @self.__room_client.on(Events.PARTICIPANT_MEDIA_STATE_CHANGED)
        def on_media_status_changed(data):
            """
            Handle event when the media status (audio/video) changes for this participant.

            Args:
                data (dict): Data containing information about the media status change.
            """
            if data["peerId"] == self.__id:
                new_state = data["newState"]
                kind = data["kind"]
                if kind == "audio":
                    self.__mic_on = new_state
                elif kind == "video":
                    self.__webcam_on = new_state
                logger.debug(f"Participant :: {self.id} :: media-status-changed")
                self.__emit(
                    VideoSDKEvents.EV_MEDIA_STATUS_CHANGED.value,
                    {
                        "peerId": data["peerId"],
                        "kind": data["kind"],
                        "newState": data["newState"],
                    },
                )

        self.__internal_event_listeners[Events.PARTICIPANT_MEDIA_STATE_CHANGED] = (
            on_media_status_changed
        )

        @self.__room_client.on(Events.VIDEO_QUALITY_CHANGED)
        def on_video_quality_changed(data):
            """
            Handle event when the video quality changes for this participant.

            Args:
                data (dict): Data containing information about the video quality change.
            """
            if data["peerId"] == self.__id:
                logger.debug(f"Participant :: {self.id} :: video-quality-changed")
                self.__emit(
                    VideoSDKEvents.EV_VIDEO_QUALITY_CHANGED.value,
                    {
                        "peerId": data["peerId"],
                        "prevQuality": data["prevQuality"],
                        "currentQuality": data["currentQuality"],
                    },
                )

        self.__internal_event_listeners[Events.VIDEO_QUALITY_CHANGED] = (
            on_video_quality_changed
        )

    def remove(self):
        """
        Remove this participant from the room.
        """
        self.__room_client.remove_peer(self.__id)

    def enable_mic(self):
        """
        Enable the microphone for this participant.
        """
        self.__room_client.enable_peer_mic(self.__id)

    def disable_mic(self):
        """
        Disable the microphone for this participant.
        """
        self.__room_client.disable_peer_mic(self.__id)

    def enable_webcam(self):
        """
        Enable the webcam for this participant.
        """
        self.__room_client.enable_peer_webcam(self.__id)

    def disable_webcam(self):
        """
        Disable the webcam for this participant.
        """
        self.__room_client.disable_peer_webcam(self.__id)

    def capture_image(
        self,
        filepath="captured_image.jpg",
        desr_height: int = None,
        desr_width: int = None,
    ):
        """
        Capture an image from the participant's video stream and save it to a file.

        Args:
            filepath (str, optional): The file path where the image will be saved. Default is "captured_image.jpg".
            desr_height (int, optional): Desired height of the captured image. Default is None.
            desr_width (int, optional): Desired width of the captured image. Default is None.
        """
        self.__tasks["capture_image"] = asyncio.ensure_future(
            self.async_capture_image(filepath, desr_height, desr_width)
        )

    async def async_capture_image(
        self, filepath="captured_image.jpg", desr_height=None, desr_width=None
    ) -> Image:
        """
        Asynchronously capture an image from the participant's video stream and return it as a PIL Image object.

        Args:
            filepath (str, optional): The file path where the image will be saved. Default is "captured_image.jpg".
            desr_height (int, optional): Desired height of the captured image. Default is None.
            desr_width (int, optional): Desired width of the captured image. Default is None.

        Returns:
            Image: The captured image as a PIL Image object.
        """
        try:
            streams = [
                stream for stream in self.__streams.values() if stream.kind == "video"
            ]

            if len(streams) == 0:
                raise Exception("Video Track Not Found")

            frame = await streams[0].track.current_frame()

            img = frame.to_ndarray(format="rgb24")

            video_height, video_width, _ = img.shape

            ratio = 16 / 9

            if desr_height is None and desr_width is None:
                desr_height = video_height
                desr_width = video_width

            if desr_height is not None:
                if desr_height <= 0 or desr_height > video_height:
                    desr_height = video_height

            if desr_width is not None:
                if desr_width <= 0 or desr_width > video_width:
                    desr_width = video_width

            if desr_height is None:
                desr_height = desr_width / ratio

            if desr_width is None:
                desr_width = desr_height * ratio

            img = Image.fromarray(img)
            resized_img = img.resize((int(desr_width), int(desr_height)))

            resized_img.save(filepath, "JPEG")
            logger.debug(f"Image captured and saved as {filepath}")

            buffered = BytesIO()

            resized_img.save(buffered, format="JPEG")

            base64_data = base64.b64encode(buffered.getvalue()).decode("utf-8")
            return base64_data

        except Exception as e:
            logger.error(f"Error in capture_image_from_stream: {e}")
            self.__room_client.emit(Events.ERROR, f"Not Able to capture image {e}")

    def _add_stream(self, producer: Producer):
        """
        Add a new stream (producer) for this participant.

        Args:
            producer (Producer): The producer object representing the stream.
        """
        stream = Stream(track=producer.track)
        self.__emit(VideoSDKEvents.EV_STREAM_ENABLED.value, stream)
        self.__streams[producer.id] = stream

    def _remove_stream(self, producer: Producer):
        """
        Remove a stream (producer) from this participant.

        Args:
            producer (Producer): The producer object representing the stream.
        """
        if producer.id in self.__streams:
            stream = self.__streams.pop(producer.id)
            self.__emit(VideoSDKEvents.EV_STREAM_DISABLED.value, stream)
            stream.track.stop()

    def _update_stream(self, producer: Producer):
        """
        Update an existing stream (producer) for this participant.

        Args:
            producer (Producer): The producer object representing the stream.
        """
        self.streams[producer.id] = producer.track

    @property
    def id(self) -> str:
        return self.__id

    @property
    def display_name(self) -> str:
        return self.__display_name

    @property
    def local(self) -> bool:
        return self.__local

    @property
    def mode(self) -> str:
        return self.__mode

    @property
    def meta_data(self) -> Any:
        return self.__meta_data

    @property
    def listeners(self) -> list[BaseEvent]:
        return self.__listeners

    @property
    def mic_on(self) -> bool:
        return self.__mic_on

    @property
    def webcam_on(self) -> bool:
        return self.__webcam_on

    @property
    def streams(self) -> dict[str, Stream]:
        return self.__streams

    @property
    def device(self) -> dict[str, Any]:
        return self.__device
